/**********************************************************************
*                                                                     *
*   Associative Coder of Buyanovsky (ACB)                             *
*                                                                     *
* acb.cpp                                                             *
*                                                                     *
* Implemented by Martin Decky, distributed under GNU GPL              *
*                                                                     *
**********************************************************************/

#include <stdexcept>
#include "acb.h"

using namespace std;
using namespace ACB;


/* ACB encoder initialization
 */
Encoder::Encoder(Vocabulary & _voc, TripletEncoder & _trip, istream & _ins) {
	trip = &_trip;
	voc = &_voc;
	ins = &_ins;
}


/* ACB encoder finalization
 *
 *  Flush any pending output.
 */
Encoder::~Encoder() {
	trip->Flush();
}


/* Encode
 *
 *  Encode up to 'cnt' characters from the input stream.
 */
const Encoder & Encoder::operator+=(const size_t cnt) {
	if (cnt == 0)
		return *this;

	size_t distance;
	bool sign;
	size_t match;
	char next;
	bool complete;
	char * buffer;
	
	buffer = (char * ) malloc(sizeof(char) * cnt);  /* Allocate buffer for input characters */
	if (buffer != NULL) {
		try {
			ins->read(buffer, cnt);                                     /* Read input stream */
			streamsize rd = ins->gcount();
			voc->Input(buffer, rd);                                     /* Add characters to look-ahead buffer */
			
			while (!voc->Exhausted()) {                                 /* Until any characters in look-ahead buffer */
				voc->Compare(distance, sign, match, next, complete);    /* Get compression triplet */
				trip->Encode(distance, sign, match, next, complete);    /* Encode triplet to output */
			}
		} catch(...) {
			free(buffer);
			throw;
		}
		free(buffer);
	} else
		throw bad_alloc();

	return *this;
}


/* ACB decoder initialization
 */
Decoder::Decoder(Vocabulary & _voc, TripletDecoder & _trip, ostream & _outs) {
	trip = &_trip;
	voc = &_voc;
	outs = &_outs;
}

/* ACB decoder finalization
 *
 *  Flush any pending output.
 */
Decoder::~Decoder() {
	outs->flush();
}

/* Decode
 *
 *  Decode 'cnt' compression triples to characters
 *  and send them to the output stream.
 */
const Decoder & Decoder::operator+=(const size_t cnt) {
	if (cnt == 0)
		return *this;

	size_t i = cnt;	
	size_t distance;
	bool sign;
	size_t match;
	char next;
	bool complete;
	
	while ((i > 0) && (trip->Decode(distance, sign, match, next, complete))) {      /* While there are any triples to decode */
		char * buffer;
		size_t size;
		
		buffer = voc->GetContent(distance, sign, match, next, complete, size);      /* Get data from the triplet */
		if (buffer != NULL) {
			try {
				outs->write(buffer, size);                                          /* Send to output */
			} catch(...) {
				free(buffer);
				throw;
			}
			free(buffer);                                                           /* Free the buffer */
		}
		i--;
	}

	return *this;
}

