/**********************************************************************
*                                                                     *
*   Associative Coder of Buyanovsky (ACB)                             *
*                                                                     *
* triplet.cpp                                                         *
*                                                                     *
* Implemented by Martin Decky, distributed under GNU GPL              *
*                                                                     *
**********************************************************************/

#include "triplet.h"

using namespace std;
using namespace ACB;

#define BUFITEMS 8


/* Basic triplet encoder initialization
 */
TripletEncoder::TripletEncoder(ostream & _outs) {
	outs = &_outs;
	
	itemsize = sizeof(buftype) * 8;
	bufsize = BUFITEMS * itemsize;
	bufpos = 0;
	
	bitbuf = new buftype[BUFITEMS];
	if (bitbuf == NULL)
		throw bad_alloc();

	for (size_t i = 0; i < BUFITEMS; i++)   /* Cleanup output buffer */
		bitbuf[i] = 0;
}

/* Basic triplet encoder finalization
 *
 *  Flush all pending data to output
 */
TripletEncoder::~TripletEncoder() {
	Flush();
	delete[] bitbuf;
}


/* Encode triplet
 */
void TripletEncoder::Encode(const size_t distance, const bool sign, const size_t match, const char next, const bool complete) {
	buftype zero = 0;
	buftype one = 1;

	if (match == 0)
		PushBits(&zero, 1);    /* Zero match characters, write just flag 0 */
	else {
		PushBits(&one, 1);
		PushBits((buftype * ) &sign, 1);
		PushBits((buftype * ) &distance, sizeof(distance) * 8);
		PushBits((buftype * ) &match, sizeof(match) * 8);
	}
	if (complete)
		PushBits(&one, 1);     /* Match is complete, write just flag 1 */
	else {
		PushBits(&zero, 1);
		PushBits((buftype * ) &next, sizeof(next) * 8);
	}
}


/* Write bits
 *
 *  Push 'cnt' bits from 'bits' to output stream (via output
 *  buffer.
 */
void TripletEncoder::PushBits(const buftype * bits, const size_t cnt) {
	for (size_t i = 0; i < cnt; i++) {
		bitbuf[bufpos / itemsize] |= ((bits[i / itemsize] >> (i % itemsize)) & 1) << (bufpos % itemsize);
		bufpos++;
		
		if (bufpos == bufsize)    /* Output buffer full, flush to output stream */
			Flush();
	}
}


/* Flush output
 *
 *  Flush all pending data to output
 */
void TripletEncoder::Flush(void) {
	size_t max = bufpos / itemsize;
	if (bufpos % itemsize > 0)   /* Count how many items really write to output */
		max++;

	for (size_t i = 0; i < max; i++) {
		outs->write((char * ) &(bitbuf[i]), sizeof(buftype));  /* Output item */
		bitbuf[i] = 0;                                         /* Cleanup item */
	}
	
	bufpos = 0;
}


/* Basic triplet decoder initialization
 */
TripletDecoder::TripletDecoder(istream & _ins) {
	ins = &_ins;
	
	itemsize = sizeof(buftype) * 8;
	bufsize = BUFITEMS * itemsize;
	buflen = 0;
	bufpos = 0;
	
	bitbuf = new buftype[BUFITEMS];
	if (bitbuf == NULL)
		throw bad_alloc();
}


/* Basic triplet decoder finalization
 */
TripletDecoder::~TripletDecoder() {
	delete[] bitbuf;
}


/* Decode triplet
 *
 *  Return 'true' if a valid triplet was decoded,
 *  'false' if the triplet is not valid or no more
 *  triplets are in the input stream.
 */	
bool TripletDecoder::Decode(size_t & distance, bool & sign, size_t & match, char & next, bool & complete) {
	buftype flag;
	
	try {
		PopBits(&flag, 1);
		if (flag == 1) {
			PopBits((buftype * ) &sign, 1);
			PopBits((buftype * ) &distance, sizeof(distance) * 8);
			PopBits((buftype * ) &match, sizeof(match) * 8);
		} else {
			sign = false;
			distance = 0;
			match = 0;
		}
	
		PopBits(&flag, 1);
		if (flag == 1)
			complete = true;
		else {
			complete = false;
			PopBits((buftype * ) &next, sizeof(next) * 8);
		}
	} catch (...) {
		return false;
	}
	
	return true;
}

/* Read bits
 *
 *  Pop 'cnt' bits from input stream (via input buffer)
 *  to 'bits'.
 */
void TripletDecoder::PopBits(buftype * bits, const size_t cnt) {
	size_t max = cnt / itemsize;
	if (cnt % itemsize > 0)              /* Count how many items in 'bits' must be cleaned up */
		max++;

	for (size_t i = 0; i < max; i++)     /* Cleanup items in 'bits' */
		bits[i] = 0;

	for (size_t i = 0; i < cnt; i++) {
		if (bufpos == buflen)            /* Input buffer is exhausted, fill in new data */
			Fill();
		
		bits[i / itemsize] |= ((bitbuf[bufpos / itemsize] >> (bufpos % itemsize)) & 1) << (i % itemsize);
		bufpos++;
	}
}

/* Fill in input buffer
 */
void TripletDecoder::Fill(void) {
	size_t i = 0;

	while ((i < BUFITEMS) && (!ins->eof())) {              /* Read at most BUFITEMS items */
		ins->read((char *) &(bitbuf[i]), sizeof(buftype));
		streamsize rd = ins->gcount();
		if (rd == (streamsize) sizeof(buftype))
			i++;                                           /* Increase fullness of the buffer on sucessful read */
	}
	buflen = i * itemsize;
	bufpos = 0;
	
	if (buflen == 0)
		throw ios_base::failure("No more data in input stream.");
}
