/**********************************************************************
*                                                                     *
*   Associative Coder of Buyanovsky (ACB)                             *
*                                                                     *
* vocabulary.cpp                                                      *
*                                                                     *
* Implemented by Martin Decky, distributed under GNU GPL              *
*                                                                     *
**********************************************************************/

#include <stdexcept>
#include "vocabulary.h"

using namespace std;
using namespace ACB;


/* Vocabulary initialization
 *
 *  Initializes empty search and look-ahead buffer
 */
Vocabulary::Vocabulary() {
	buffer = NULL;
	bufsize = 0;
	sbuf = 0;
}


/* Vocabulary Finalization
 *
 *  Deallocates buffers
 */
Vocabulary::~Vocabulary() {
	free(buffer);
}


/* Look-ahead buffer empty
 *
 *  Returns 'true' if the look-ahead buffer is empty
 *  (no more characters to proceed)
 */
bool Vocabulary::Exhausted(void) const {
	return (sbuf >= bufsize);
}


/* Put new characters to look-ahead buffer
 *
 *  Adds 'cnt' characters pointed by 'input' to look-ahead
 *  buffer
 */
void Vocabulary::Input(const char * input, const size_t cnt) {
	if (cnt > 0) {
		buffer = (char * ) realloc(buffer, sizeof(char) * (bufsize + cnt));    /* Reallocate current buffers */
		if (buffer == NULL)
			throw bad_alloc();
		memcpy(buffer + bufsize, input, sizeof(char) * cnt);                   /* Copy characters from 'input' at the end of 'buffer' */
		bufsize += cnt;                                                        /* Adjust buffers sizes */
	}
}


/* Get new compression triplet
 *
 *  Computes the best context/content match against current
 *  vocabulary, updates the vocabulary and search buffer
 *  according the match.
 */
void Vocabulary::Compare(size_t & distance, bool & sign, size_t & match, char & next, bool & complete) {
	if ((bufsize == 0) || (Exhausted()))
		throw domain_error("No characters to process in " + string(__FUNCTION__));
	
	if (sbuf == 0) {         /* Search buffer is empty */
		distance = 0;
		sign = false;
		match = 0;
		next = buffer[0];    /* Copy first character from look-ahead buffer to search buffer */
		complete = false;
		sbuf++;
		return;              /* Return the first character as single triplet with no match */
	}

	size_t cxmatch = ContextMatch();    /* Get current context (search buffer) match in vocabulary */

	match = 0;
	next = buffer[sbuf];
	complete = false;
	iterator it;
	size_t i;
	size_t ctmatch = 0;
	
	for (it = list.begin(), i = 0; it != list.end(); it++, i++) {     /* Get the best content match in vocabulary */
		char curnext;
		bool curcomplete;
		size_t curmatch = ContentMatch(*it, curnext, curcomplete);
		
		if (curmatch > match) {
			match = curmatch;
			next = curnext;
			complete = curcomplete;
			ctmatch = i;
		}
	}
	
	if (cxmatch <= ctmatch) {      /* Compute the distance between context and content match */
		distance = ctmatch - cxmatch;
		sign = false;
	} else {
		distance = cxmatch - ctmatch;
		sign = true;
	}
	
	if (complete)                  /* Updates the vocabulary according the match */
		UpdateVocabulary(match);
	else
		UpdateVocabulary(match + 1);
}


/* Get data from compression triplet
 *
 *  Using current vocabulary, gets the matching data
 *  according context match and 'distance'. Updates
 *  the vocabulary according the match.
 */
char * Vocabulary::GetContent(const size_t distance, const bool sign, const size_t match, const char next, const bool complete, size_t & size) {
	char * output;
	
	if (sbuf == 0) {                                         /* Search buffer is empty */
		if ((match == 0) && (!complete)) {                   /* The first compression triplet must code a single character 'next' */
			Input(&next, 1);                                 /* Put 'next' to search buffer */
			sbuf++;
			
			size = 1;
			output = (char * ) malloc(sizeof(char) * size);
			if (output == NULL)
				throw bad_alloc();
			memcpy(output, &next, sizeof(char));
				
			return output;
		}
		return NULL;
	}

	if (match > 0) {
		iterator it;
		size_t i;
		size_t pred = ContextMatch();                        /* Get context match according vocabulary */

		if (sign)                                            /* Compute content match from context match and 'distance' */
			pred -= distance;
		else
			pred += distance;
	
		for (it = list.begin(), i = 0; it != list.end(); it++, i++) {       /* Get exactly 'match' characters from content indexed by 'pred' */
			if (i == pred) {
				if (complete) {                                             /* No final character */
					size = match;
					output = (char * ) malloc(sizeof(char) * size);
					if (output == NULL)
						throw bad_alloc();
					memcpy(output, buffer + (*it), sizeof(char) * match);
					
					Input(output, size);                                    /* Add the characters to search buffer */
					UpdateVocabulary(size);                                 /* Update vocabulary with the characters */
					
					return output;
				} else {                                                    /* Same as above, but with final character 'next' */
					size = match + 1;
					output = (char * ) malloc(sizeof(char) * size);
					if (output == NULL)
						throw bad_alloc();
					memcpy(output, buffer + (*it), sizeof(char) * match);
					memcpy(output + match, &next, sizeof(char));
					
					Input(output, size);
					UpdateVocabulary(size);
					
					return output;
				}
			}
		}
	}
	
	if (!complete) {                                         /* Compression triplet codes just a single character 'next' */
		size = 1;
		output = (char * ) malloc(sizeof(char) * size);
		if (output == NULL)
			throw bad_alloc();
		memcpy(output, &next, sizeof(char));
		
		Input(output, size);                                 /* Add 'next' to search buffer */
		UpdateVocabulary(size);                              /* Update vocabulary with 'next' */
				
		return output;
	}
			
	return NULL;
}


/* Get context match
 *
 *  Return the position in vocabulary where
 *  should be the current context sorted in.
 */
size_t Vocabulary::ContextMatch(void) {
	iterator it;
	size_t i;
	
	for (it = list.begin(), i = 0; it != list.end(); it++, i++) {
		if (ContextCompare(*it))        /* Currect context (reversed) is lexically smaller than (reversed) entry in vocabulary */
			return i;
	}
	
	return list.size();                 /* Current context is greater than all entries -> goes at the end of the vocabulary */
}


/* Get length of content match
 *
 *  Returns the length of look-ahead and content match.
 */
size_t Vocabulary::ContentMatch(const size_t shift, char & next, bool & complete) const {
	size_t baselen = bufsize - sbuf;
	size_t len = sbuf - shift;

	for (size_t i = 0; i < baselen; i++)                              /* Compare look-ahead buffer with content defined by 'shift' */
		if ((i >= len) || (buffer[sbuf + i] != buffer[shift + i])) {
			complete = false;
			next = buffer[sbuf + i];                                  /* First non-matching character */
			return i;
		}
		
	complete = true;           /* Look-ahead and content match completely, no non-matching character */
	return min(baselen, len);
}


/* Update vocabulary
 *
 *  Inserts the contexts of 'cnt' characters from look-ahead
 *  buffer to their proper positions in the vocabulary.
 */
void Vocabulary::UpdateVocabulary(const size_t cnt) {
	for (size_t i = cnt; (i > 0) && (sbuf < bufsize); i--) {
		bool inserted = false;
	
		for (iterator it = list.begin(); it != list.end(); it++) {
			if (ContextCompare(*it)) {      /* Currect context (reversed) is lexically smaller than (reversed) entry in vocabulary */
				list.insert(it, sbuf);
				inserted = true;
				break;
			}
		}
	
		if (!inserted)
			list.push_back(sbuf);           /* Current context is greater than all entries -> goes at the end of the vocabulary */
		
		sbuf++;                             /* Enlarge search buffer (thus shorten look-ahead buffer) */
	}

#ifdef DEBUG
	DebugPrint();
#endif

}


/* Compare context and search buffer
 *
 *  Do a lexical comparation (from right to left) between 
 *  current search buffer and context defined by 'shift'.
 */
bool Vocabulary::ContextCompare(const size_t shift) const {
	if (shift >= bufsize)
		throw out_of_range("shift out of range in " + string(__FUNCTION__));
		
	size_t i = shift;
	size_t j = sbuf;
	
	while ((i > 0) && (j > 0)) {
		i--;
		j--;
		
		if (buffer[i] > buffer[j])
			return true;
		else if (buffer[i] < buffer[j])
			return false;
	}
	
	if (shift > sbuf)
		return true;

	return false;
}


/* Display current vocabulary
 *
 *  Prints current vocabulary to stderr.
 */
void Vocabulary::DebugPrint(void) {
	iterator it;
	size_t i;
	size_t max = 0;
	
	for (it = list.begin(); it != list.end(); it++)
		if (*it > max)
			max = *it;
			
	for (it = list.begin(), i = 0; it != list.end(); it++, i++) {
	    cerr.width(5);
		cerr << right << i;
		cerr.width(0);
		cerr << left << ": ";
		
		cerr.width(sbuf - *it);
		cerr << right << "";
		cerr.width(0);
		cerr << left;
		cerr.write(buffer, *it);
		
		cerr << "|";
		cerr.write(buffer + (*it), sbuf - (*it));
		cerr << endl;
	}
	cerr << endl;
}
